# -*- coding: utf-8 -*-
from typing import Callable


class Infix:
    """
    Implements a custom Infix operator using  the 
    operators '<<', '>>' and '|'.

    Examples
    --------
    >>> x = Infix(lambda x, y: x * y)
    >>> print(2 | x | 4)
    8

    >>> x = Infix(lambda x, y: x + y)
    >>> print(2 << x >> 4)
    6
    """

    def __init__(self, function: Callable):
        self.function = function

    def __ror__(self, other):
        return Infix(lambda x, self=self, other=other: self.function(other, x))

    def __or__(self, other):
        return self.function(other)

    def __rlshift__(self, other):
        return Infix(lambda x, self=self, other=other: self.function(other, x))

    def __rshift__(self, other):
        return self.function(other)

    def __call__(self, value1, value2):
        return self.function(value1, value2)


# -*- coding: utf-8 -*-
import time
import sys
from typing import Callable
from ..typing import issequence


__all__ = ['float_to_str_sig', 'floatformatter', 'issequence', 'suppress']


def floatformatter(*args, sig: int=6, **kwargs) -> str:
    """
    Returns a formatter, which essantially a string temapate
    ready to be formatted.
    
    Parameters
    ----------
    sig : int, Optional
        Number of significant digits. Default is 6.
        
    Returns
    -------
    string
        The string to be formatted.
    
    Examples
    --------    
    >>> from dewloosh.core import DeepDict
    >>> data = DeepDict()
    >>> data['a']['b']['c']['e'] = 1
    >>> data['a']['b']['d'] = 2
    >>> data.containers()
            
    """
    return "{" + "0:.{}g".format(sig) + "}"


def float_to_str_sig(value, *args, sig: int=6, atol: float=1e-7, 
                     **kwargs) -> str:
    """
    Returns a string representation of a floating point number, with
    given significant digits.

    Parameters
    ----------
    value : float or a sequence of floats
        A single value, or an iterable.

    sig : int
        Number of significant digits.

    atol : float
        Floating point tolerance. Values smaller than this 
        in the absolute sense are treated as zero.

    Returns
    -------
    string or a sequence of strings
        String representation of the provided input.
    
    Example
    --------
    Print the value of pi as a string with 4 significant digits:
    
    >>> from dewloosh.core.tools import float_to_str_sig
    >>> import math
    >>> float_to_str_sig(math.pi, sig=4)
    '3.142'
    
    """
    if not issequence(value):
        if atol is not None:
            if abs(value) < atol:
                value = 0.0
        return floatformatter(sig=sig).format(value)
    else:
        try:
            import numpy as np
        except ImportError:
            raise ImportError("You need numpy for this.")
        value = np.array(value)
        if atol is not None:
            inds = np.where(np.abs(value) < atol)[0]
            value[inds] = 0.0
        formatter = floatformatter(sig=sig)
        def f(v): return formatter.format(v)
        return list(map(f, value))


def timeit(fnc : Callable) -> float:
    """
    A simple decorator to measure execution time of a function.
    """
    def inner(*args, **kwargs):
        t0 = time.time()
        fnc(*args, **kwargs)
        t1 = time.time()
        return t1-t0
    return inner


def suppress(fnc: Callable) -> Callable:
    """
    Decorator that wraps a function to suppress it's calls to `print`.
    """
    def inner(*arg):
        original_stdout = sys.stdout
        sys.stdout = None
        res = fnc(*arg)
        sys.stdout = original_stdout
        return res
    return inner


"""Downloadable datasets collected from various sources.

Once downloaded, these datasets are stored locally allowing for the
rapid reuse of these datasets.

Examples
--------
>>> from dewloosh.core.downloads import download_stand
>>> download_stand()
...

"""

from functools import partial
import os
import shutil
from urllib.request import urlretrieve
import zipfile

try:
    import pyvista
    __haspv__ = True
except ImportError:
    __haspv__ = False

from . import EXAMPLES_PATH, DEWLOOSH_DATA_PATH as DATA_PATH


def _check_examples_path():
    """Check if the examples path exists."""
    if not EXAMPLES_PATH:
        raise FileNotFoundError(
            'EXAMPLES_PATH does not exist.  Try setting the '
            'environment variable `DEWLOOSH_USERDATA_PATH` '
            'to a writable path and restarting python'
        )


def delete_downloads():
    """Delete all downloaded examples to free space or update the files.

    Returns
    -------
    bool
        Returns ``True``.

    Examples
    --------
    Delete all local downloads.

    >>> from dewloosh.core import delete_downloads
    >>> delete_downloads()  # doctest:+SKIP
    True

    """
    _check_examples_path()
    shutil.rmtree(EXAMPLES_PATH)
    os.makedirs(EXAMPLES_PATH)
    return True


def _decompress(filename):
    _check_examples_path()
    zip_ref = zipfile.ZipFile(filename, 'r')
    zip_ref.extractall(EXAMPLES_PATH)
    return zip_ref.close()


def _get_vtk_file_url(filename):
    return f'https://github.com/dewloosh/dewloosh-data/raw/main/{filename}'


def _http_request(url):
    return urlretrieve(url)


def _repo_file_request(repo_path, filename):
    return os.path.join(repo_path, 'Data', filename), None


def _retrieve_file(retriever, filename):
    """
    Retrieve file and cache it in dewloosh.core.EXAMPLES_PATH.

    Parameters
    ----------
    retriever : str or callable
        If str, it is treated as a url.
        If callable, the function must take no arguments and must
        return a tuple like (file_path, resp), where file_path is
        the path to the file to use.
    filename : str
        The name of the file.
    
    Notes
    -----
    You must have `PyVista` installed to handle zip files.
    
    """
    _check_examples_path()
    # First check if file has already been downloaded
    local_path = os.path.join(EXAMPLES_PATH, os.path.basename(filename))
    local_path_no_zip = local_path.replace('.zip', '')
    if os.path.isfile(local_path_no_zip) or os.path.isdir(local_path_no_zip):
        return local_path_no_zip, None
    if isinstance(retriever, str):
        retriever = partial(_http_request, retriever)
    saved_file, resp = retriever()
    # new_name = saved_file.replace(os.path.basename(saved_file), os.path.basename(filename))
    # Make sure folder exists!
    if not os.path.isdir(os.path.dirname((local_path))):
        os.makedirs(os.path.dirname((local_path)))
    if DATA_PATH is None:
        shutil.move(saved_file, local_path)
    else:
        if os.path.isdir(saved_file):
            shutil.copytree(saved_file, local_path)
        else:
            shutil.copy(saved_file, local_path)
    if __haspv__:
        if pyvista.get_ext(local_path) in ['.zip']:
            _decompress(local_path)
            local_path = local_path[:-4]
    return local_path, resp


def _download_file(filename):
    if DATA_PATH is None:
        url = _get_vtk_file_url(filename)
        retriever = partial(_http_request, url)
    else:
        if not os.path.isdir(DATA_PATH):
            raise FileNotFoundError(
                f'Data repository path does not exist at:\n\n{DATA_PATH}'
            )
        if not os.path.isdir(os.path.join(DATA_PATH, 'Data')):
            raise FileNotFoundError(
                f'Data repository does not have "Data" folder at:\n\n{DATA_PATH}'
            )
        retriever = partial(_repo_file_request, DATA_PATH, filename)
    return _retrieve_file(retriever, filename)


def _download_and_read(filename):
    saved_file, _ = _download_file(filename)
    return saved_file

###############################################################################


def download_stand():  # pragma: no cover
    """
    Downloads a tetrahedral mesh of a stand in vtk format.

    Returns
    -------
    str
        A path to a file on your filesystem.

    Example
    --------
    >>> from dewloosh.core.downloads import download_stand
    >>> download_stand()
    ...
    
    """
    return _download_file('stand.vtk')[0]



# -*- coding: utf-8 -*-
from .meta import *


__all__ = ["ABC_Strong", "ABC_Weak", "ABC_Safe"]


class ABC_Weak(metaclass=ABCMeta_Weak):
    """
    Helper class that provides a standard way to create an ABC using
    inheritance.
    """
    __slots__ = ()


class ABC_Strong(metaclass=ABCMeta_Strong):
    """Helper class that provides a standard way to create an ABC using
    inheritance.
    """
    __slots__ = ()


class ABC_Safe(metaclass=ABCMeta_Safe):
    """
    Helper class that provides a standard way to create an ABC using
    inheritance.
    """
    __slots__ = ()




# -*- coding: utf-8 -*-
from abc import ABCMeta


__all__ = ["ABCMeta_Weak", "ABCMeta_Strong", "ABCMeta_Safe"]


def _is_callable(n, v): return callable(v) and ('__' not in n)


class ABCMeta_Weak(ABCMeta):
    """
    Standard python metaclass. It follows weak abstraction in the meaning, that
    it is enough to implement the abstarcts in the instance, they impose no
    restriction on the inherited class itself. Therefore, error only occurs at
    runtime.
    """

    @staticmethod
    def _get_cls_methods(namespace, nomagic=False):
        """
        Returns the callable items in the namespace of the class.
        If `nomagic=False`, magic functions with trailing double 
        underscores are not returned.
        """
        if not nomagic:
            return {name for name, value in namespace.items()
                    if callable(value)}
        else:
            return {name for name, value in namespace.items()
                    if _is_callable(name, value)}
            
    @staticmethod
    def _get_cls_abstracts(namespace):
        """
        Returns the callable items in the namespace of the class.
        If `nomagic=False`, magic functions with trailing double 
        underscores are not returned.
        """            
        return getattr(namespace, "__abstractmethods__", set())
            
    @staticmethod
    def _get_base_methods(bases):
        """
        Returns the callable items in the namespace of the class.
        If `nomagic=False`, magic functions with trailing double 
        underscores are not returned.
        """
        base_methods = {}
        for base in bases:
            base_methods[base.__name__] = {}
            for k, v in base.__dict__.items():
                if _is_callable(k, v):
                    base_methods[base.__name__][k] = v
        return base_methods
                
    @staticmethod
    def _get_base_abstracts(bases):
        """
        Returns the callable items in the namespace of the class.
        If `nomagic=False`, magic functions with trailing double 
        underscores are not returned.
        """
        base_abc_methods = {}
        for base in bases:
            base_abc_methods[base.__name__] = \
                getattr(base, "__abstractmethods__", set())
        return base_abc_methods


class ABCMeta_Strong(ABCMeta_Weak):
    """
    Strong Python metaclass. It follows strong abstraction in the meaning, that
    an abstract function of any of the base classes must be implemented or
    delayed with the use of another @abstractmethod decorator.
    Error occurs at compilation time.
    """

    def __init__(self, name, bases, namespace, *args, **kwargs):
        super().__init__(name, bases, namespace, *args, **kwargs)

    def __new__(metaclass, name, bases, namespace, *args, **kwargs):
        cls = super().__new__(metaclass, name, bases, namespace, *args,
                              **kwargs)
        cls_methods = metaclass._get_cls_methods(namespace)
        for base in bases:
            base_abstracts = set()
            for name in getattr(base, "__abstractmethods__", set()):
                value = getattr(cls, name, None)
                if getattr(value, "__isabstractmethod__", False):
                    base_abstracts.add(name)
            for abstract in base_abstracts:
                if abstract not in cls_methods:
                    err_str = """Can't create abstract class {classname}!
                    {classname} must implement abstract method {method} of
                    class {base_class}!""".format(
                        classname=name,
                        method=abstract,
                        base_class=base.__name__)
                    raise TypeError(err_str)
        return cls


class ABCMeta_Safe(ABCMeta_Weak):
    """
    Python metaclass for safe inheritance. Throws a TypeError
    if a method tries to shadow a definition in any of the base
    classes.
    """

    def __init__(self, name, bases, namespace, *args, **kwargs):
        super().__init__(name, bases, namespace, *args, **kwargs)

    def __new__(metaclass, name, bases, namespace, *args, **kwargs):
        cls = super().__new__(metaclass, name, bases, namespace, *args,
                              **kwargs)
        cls_methods = metaclass._get_cls_methods(namespace, nomagic=True)
        for base in bases:
            for method in cls_methods:
                if hasattr(base, method):
                    err_str = """Can't create abstract class {classname}!
                        Method {method} is already implemented in class
                        {base_class}!""".format(
                        classname=name,
                        method=method,
                        base_class=base.__name__)
                    raise TypeError(err_str)
        return cls


# -*- coding: utf-8 -*-
"""
Multiple solutions on how to enforce an abstract class property
on an object. Generic types like 'List[int]' are not allowed, 
because the don't work well with the isinstance()-like methods.
"""

__all__ = ['abstract_class_property', 'setproperty']


def abstract_class_property(**kwargs):
    return abstract_class_property_B(**kwargs)


def setproperty(**kwargs):
    def decorator(cls):
        for key, value in kwargs.items():
            setattr(cls, key, value)
        return cls
    return decorator


"""
Implementation A : class annotations solution using an object descriptor 
and a wrapper class with classic pythonic manipulations.
"""


def abstract_class_property_A(**kwargs):
    """
    Decorator function to decorate objects with abstract
    class properties. Leaves behind another decorator 
    that takes a class as its input.
    """

    def abstractor(WrappedClass):

        class PropertyWrapper(WrappedClass):
            """
            A Wrapper class to decorate objects with abstract class properties.
            The class has a default dummy annotation, just so that we can append 
            the items of the input dictionary d to the class definition.
            The dummy property is deleted at the end.
            """
            _dummy_: None

            def __init__(self, *args, **kwargs):
                WrappedClass.__init__(self)
                d_ = dict()
                d_props = PropertyWrapper.__dict__.get('__annotations__', {})
                for key, value in d_props.items():
                    d_[key] = value
                d_self = self.__class__.__dict__.get('__annotations__', {})
                for key, value in d_self.items():
                    d_[key] = value
                for key in d_.keys():
                    if not hasattr(self, key):
                        raise AttributeError(f'required attribute {key} not present '
                                             f'in {self.__class__}')
                return

        res = PropertyWrapper
        d_ = res.__dict__['__annotations__']
        for key, value in kwargs.items():
            d_[key] = value
        for key, value in WrappedClass.__dict__.get('__annotations__', {}).items():
            d_[key] = value
        del d_['_dummy_']
        return res

    return abstractor


"""
Implementation B : class annotations solution using an object descriptor and a 
wrapper class using the __mro__ property of the class.
This implementation also checks the validity of the declaration.
"""


def abstract_class_property_B(**kwargs):
    """
    Decorator function to decorate objects with abstract
    class properties. Leaves behind another decorator
    that takes a class as its input.
    """

    def abstractor(WrappedClass):

        class PropertyWrapper(WrappedClass):
            """
            A Wrapper class to decorate objects with abstract class properties.
            The class has a default dummy annotation, just so that we can append 
            the items of the input dictionary d to the class definition.
            The dummy property is deleted at the end.
            """
            _dummy_: None

            def __init__(self, *args, **kwargs):
                WrappedClass.__init__(self)
                self.__check_absclsprops__()
                return

            @classmethod
            def __absclsprops__(cls):
                """
                Collects the abstracts class properties and their
                expected types based on the MRO of the class.
                """
                t = cls.__mro__
                l = [ti.__dict__.get('__annotations__', {}) for ti in t]
                res = dict()
                for d in l:
                    for key, value in d.items():
                        res[key] = value
                return res

            def __check_absclsprops__(self):
                """
                Checks if the instance has attributes according to
                the anstract annotations. Returns True if every attribute is
                a type-correct declaration.
                """
                props = self.__class__.__absclsprops__()
                for key, value in props.items():
                    if not hasattr(self, key):
                        raise AttributeError(f'required attribute {key} not present '
                                             f'in {self.__class__}')
                    else:
                        if not isinstance(getattr(self, key), value):
                            raise TypeError(
                                'TypeError. key : {}, value = {}'.format(key, value))
                return True

        res = PropertyWrapper
        d_ = res.__dict__['__annotations__']
        for key, value in kwargs.items():
            d_[key] = value
        del d_['_dummy_']
        return res

    return abstractor